# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rest_framework import serializers
from rest_framework.reverse import reverse


class NestedHyperlinkedIdentityField(serializers.HyperlinkedIdentityField):

    def __init__(self, **kwargs):
        self.parents_query_lookups = kwargs.pop('parents_query_lookups')
        super(NestedHyperlinkedIdentityField, self).__init__(**kwargs)

    def get_url(self, obj, view_name, request, format):
        if obj.pk is None:
            return None
        lookup_value = getattr(obj, self.lookup_field)
        kwargs = {self.lookup_url_kwarg: lookup_value}
        for lookup_url_kwarg, lookup_field in self.parents_query_lookups.items():
            lookup_value = getattr(obj, lookup_field)
            lookup_kwarg = "parent_lookup_{}".format(lookup_url_kwarg)
            kwargs.update({lookup_kwarg: lookup_value})
        return reverse(view_name, kwargs=kwargs, request=request, format=format)





# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from collections import OrderedDict
from rest_framework.serializers import Serializer


class DataFrameSerializer(Serializer):

    def to_representation(self, instance):
        data = instance.to_dict('records')
        items = [
            ('count', len(data)),
            ('next', None),
            ('previous', None),
            ('results', data),
        ]
        return OrderedDict(items)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function

from rest_framework.settings import api_settings
from rest_framework.utils import html
from rest_framework.exceptions import ValidationError


class ActionExcludesModelFieldsMixin(object):

    def get_field_names(self, declared_fields, info):
        fields = list(super(ActionExcludesModelFieldsMixin, self).get_field_names(declared_fields, info))
        action_excludes = getattr(self.Meta, 'action_excludes', {}).get(getattr(self.context.get('view'), 'action', None), [])
        return [field for field in fields if field not in action_excludes]


class LenientBulkCreateSerializerMixin(object):

    def to_internal_value(self, data):
        """
        List of dicts of native values <- List of dicts of primitive datatypes.
        """
        if html.is_html_input(data):
            data = html.parse_html_list(data)

        if not isinstance(data, list):
            message = self.error_messages['not_a_list'].format(input_type=type(data).__name__)
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: [message]})

        ret = []
        errors = []

        for item in data:
            try:
                validated = self.child.run_validation(item)
            except ValidationError as exc:
                errors.append({"errors": exc.detail, "item": item})
            else:
                ret.append(validated)

        self.bulk_validation_errors = errors
        return ret



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rest_framework.relations import HyperlinkedIdentityField


class HyperlinkedIdentityWithParamsField(HyperlinkedIdentityField):

    def __init__(self, view_name=None, **kwargs):
        self.view_kwargs = kwargs.pop('view_kwargs', {})
        super(HyperlinkedIdentityWithParamsField, self).__init__(view_name, **kwargs)

    def get_url(self, obj, view_name, request, format):
        if obj.pk is None:
            return None
        lookup_value = getattr(obj, self.lookup_field)
        kwargs = {self.lookup_url_kwarg: lookup_value}
        kwargs.update(**self.view_kwargs)
        return self.reverse(view_name, kwargs=kwargs, request=request, format=format)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from pandas import json
from django.conf import settings
from django.utils import six
from rest_framework.parsers import BaseParser
from rest_framework.exceptions import ParseError


class LineDelimitedJSONParser(BaseParser):
    """
    Parses Line Delimited JSON-serialized data.
    """

    media_type = 'application/x-ldjson'

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Parses the incoming bytestream as Line Delimited JSON and returns the resulting data.

        """
        parser_context = parser_context or {}
        encoding = parser_context.get('encoding', settings.DEFAULT_CHARSET)
        return [self.parse_line(line, encoding) for line in stream]

    def parse_line(self, line, encoding):
        try:
            return json.loads(line.decode(encoding))
        except ValueError as exc:
            raise ParseError('Line delimited JSON parse error - %s' % six.text_type(exc))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from collections import defaultdict
from copy import deepcopy
from itertools import groupby

import re
from dateutil.parser import parse
from django.core.exceptions import ImproperlyConfigured

from django.utils import six

from elasticsearch_dsl.filter import F
from elasticsearch_dsl.query import FunctionScore

from rest_framework import filters
from rest_framework.exceptions import ParseError
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings

from ..filters import FilterOptionsMixin


def _function_score_params(origin_parser=None, offset_parser=None, scale_parser=None, decay_parser=None, global_disallowed=None):
    global_disallowed = global_disallowed or []
    parsers = {"origin": origin_parser or (lambda values, value: value),
               "offset": offset_parser or (lambda values, value: value),
               "scale": scale_parser or (lambda values, value: value),
               "decay": decay_parser or (lambda values, value: float(value) if value else None)}

    def wrapper(base_param, defaults=None, disallowed=None):
        disallowed = global_disallowed + (disallowed or [])
        defaults = defaults or {}
        params = {"origin": base_param,
                  "offset": "{}.offset".format(base_param),
                  "scale": "{}.scale".format(base_param),
                  "decay": "{}.decay".format(base_param)}

        def get_kwargs(request, view):
            if params["origin"] not in request.query_params and "origin" not in defaults:
                # origin has to be given via request or defaults
                return

            values = {"origin": request.query_params.get(params["origin"]) if "origin" not in disallowed else None,
                      "offset": request.query_params.get(params["offset"]) if "offset" not in disallowed else None,
                      "scale": request.query_params.get(params["scale"]) if "scale" not in disallowed else None,
                      "decay": request.query_params.get(params["decay"]) if "decay" not in disallowed else None}

            overrides = {}
            for arg, param in params.items():
                try:
                    val = parsers[arg](values, values[arg])
                    if isinstance(val, Exception):
                        raise val
                    if val is None:
                        continue
                    overrides.update({arg: val})
                except (ValueError, IndexError):
                    raise ParseError("Invalid value for '{}'".format(param))

            result = deepcopy(defaults)
            result.update(overrides)
            return result

        return get_kwargs
    return wrapper


class SearchMixin(object):

    search_param = getattr(api_settings, 'SEARCH_PARAM')

    def get_search(self, request):
        return request.query_params.get(self.search_param, "").strip()


class QueryParamFilterBuilderMixin(object):

    use_canonicals = True
    # List valid url operator
    valid_operators = []
    # Maps url operators to actual operator to use in the filter
    real_operators = {}
    # Fields used to store the value to match (should be exclusive of each other)
    value_operators = []
    # Operator negating the filter (should be one of value operator)
    negation_operators = []

    def get_fields(self, view):
        raise NotImplementedError("Subclass must implement 'get_fields'")

    def get_valid_fields(self, fields, include_canonicals=True):
        return (fields if include_canonicals else []) + ['{}.{}'.format(v, r) for v in fields for r in self.valid_operators]

    def get_operator(self, operator):
        return self.real_operators.get(operator, operator)

    def map_field(self, request, view, field):
        return field

    def map_value(self, request, view, field, operator, value):
        return value

    def parse_value(self, request, view, field, operator, value):
        return self.map_value(request, view, field, operator, value)

    def build_filter(self, field, **params):
        raise NotImplementedError("Subclass must implement 'build_filter'")

    def parse_param(self, param_name):
        parts = param_name.split('.')
        if parts[-1] in self.valid_operators:
            return '.'.join(parts[0:-1]), parts[-1]
        else:
            return param_name, None

    def get_mapped_params(self, request, view, valid_params):
        for param_name, param_value in valid_params:
            field, operator = self.parse_param(param_name)
            field = self.map_field(request, view, field)
            yield '.'.join(filter(None, [field, operator])), param_value

    def get_filter_params(self, request, view):
        valid_fields = self.get_valid_fields(self.get_fields(view), include_canonicals=self.use_canonicals)
        valid_params = list((k, v) for k, v in request.query_params.items() if k in valid_fields)
        clean_params = self.get_mapped_params(request, view, valid_params)
        field_groups = groupby(clean_params, key=lambda item: self.parse_param(item[0])[0])
        field_params = defaultdict(dict)
        for field, group in field_groups:
            for k, v in group:
                try:
                    field, operator = self.parse_param(k)
                except ValueError:
                    field, operator = k, None
                op = self.get_operator(operator)
                if op and op not in self.negation_operators:
                    field_params[field][op] = self.parse_value(request, view, field, op, v)
                else:
                    if op in self.negation_operators:
                        field_params[field]["negate"] = True
                    field_params[field][field] = self.parse_value(request, view, field, op, v)
        return field_params

    def get_filters(self, filter_params):
        filters = []
        for field, params in filter_params.items():
            negate = params.pop("negate", False)
            if self.value_operators and field not in params:
                raise ParseError("No value to match for '{}'".format(field))
            f = self.build_filter(field, **params)
            filters.append(~f if negate else f)
        return filters

    def filter_queryset(self, request, queryset, view):
        for filter in self.get_filters(self.get_filter_params(request, view)):
            queryset = queryset.filter(filter)
        return queryset


class BaseSearchFilterBackend(SearchMixin, FilterOptionsMixin, BaseFilterBackend):

    @property
    def search_type(self):
        raise NotImplementedError()

    def filter_queryset(self, request, queryset, view):
        search = self.get_search(request)
        if search:
            return queryset.query({self.search_type: dict(query=search, **self.get_options(view))})
        return queryset


class SimpleQueryStringFilter(BaseSearchFilterBackend):
    """
    Filter adding a simple query string to a search

    For the available view options, see:
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-simple-query-string-query.html

    Example::

        class YourView(APIView):

            simple_query_string_filter_options = {
                "analyzer": "snowball",
                "fields": ["body^5", "_all"],
                "default_operator": "and"
            }

    Note: Make sure to add _score to your default ordering if you want results sorted by relevance

    """
    search_type = "simple_query_string"


class MultiMatchFilter(BaseSearchFilterBackend):
    """
    Filter adding a multi match to a search

    For the available view options, see:
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-multi-match-query.html

    Example::

        class YourView(APIView):

            multi_match_options = {
                "type": "best_fields",
                "fields": ["subject", "message"],
                "tie_breaker": 0.3
            }

    Note: Make sure to add _score to your default ordering if you want results sorted by relevance

    """
    search_type = "multi_match"


class FunctionScoreFilter(FilterOptionsMixin, BaseFilterBackend):
    """
    Filter adding a function score to a search

    For the available view options, see:
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-function-score-query.html

    Example::

        class YourView(APIView):

            function_score_options = {
                "functions": [
                    {
                        "field": "geo_point",
                        "function": "gauss",
                        "kwargs": FunctionScoreFilter.geo_params("nearby", {"offset": "1km", "scale": "1km"}, disallowed=["decay"]),
                    }
                ]
            }

    kwargs can either take a dictionary, or a callable(request, view) returning a dictionary.
    The score function is ignored if the callable returns None. The request is malformed if the callable raises a ParseError

    Available param function shortcuts:

    * geo_params(param, defaults, disallowed) (e.g. &param=-36.8471855,174.7700463&param.offset=2km&param.scale=500m)
    * number_params(param, defaults, disallowed) (e.g. &param=10&param.offset=5&param.scale=2)
    * price_params(param) (e.g. &param=5)

    Note: Make sure to add _score to your default ordering if you want results sorted by relevance

    """
    geo_params = staticmethod(_function_score_params(origin_parser=lambda values, value: dict(zip(('lat', 'lon'), map(lambda v: float(v.strip()), value.split(',', 1)))),
                                                     offset_parser=lambda values, value: value if value is None or re.match("\d+[kc]?m", value) else ValueError(),
                                                     scale_parser=lambda values, value: value if value is None or re.match("\d+[kc]?m", value) else ValueError()))

    date_params = staticmethod(_function_score_params(origin_parser=lambda values, value: value,
                                                     offset_parser=lambda values, value: value if value is None or re.match("\d+d", value) else ValueError(),
                                                     scale_parser=lambda values, value: value if value is None or re.match("\d+d", value) else ValueError()))

    number_params = staticmethod(_function_score_params(origin_parser=lambda values, value: float(value),
                                                       offset_parser=lambda values, value: float(value),
                                                       scale_parser=lambda values, value: float(value)))

    price_params = staticmethod(_function_score_params(origin_parser=lambda values, value: float(value) / 2,
                                                       offset_parser=lambda values, value: float(values['origin']) / 2,
                                                       scale_parser=lambda values, value: float(values['origin']) / 3,
                                                       global_disallowed=['offset', 'scale', 'decay']))

    def get_functions(self, request, view):
        functions = []
        for f in self.get_option(view, "functions"):
                kwargs = f.get('kwargs')
                if callable(kwargs):
                    kwargs = kwargs(request, view)
                if not kwargs:
                    continue
                function = {f.get("function"): {f.get("field"): kwargs}}
                if f.get("weight", False):
                    function.update({'weight': f.get("weight")})
                functions.append(function)
        return functions

    def filter_queryset(self, request, queryset, view):
        functions = self.get_functions(request, view)
        if functions:
            params = {k: v for k, v in self.get_options(view).items() if k not in ['functions']}
            return queryset.query(FunctionScore(functions=functions, **params))
        return queryset


class TermFilter(BaseFilterBackend):

    def get_valid_terms_filters(self, view):
        filters = getattr(view, 'term_filters', [])
        if isinstance(filters, six.string_types):
            return (filters,)
        return filters

    def get_term_filters(self, request, valid_filters):
        return {k: v.strip() for k, v in request.query_params.items() if k in valid_filters}

    def filter_queryset(self, request, queryset, view):
        valid_filters = self.get_valid_terms_filters(view)

        term_filters = self.get_term_filters(request, valid_filters)

        for field, values in term_filters.items():
            queryset = queryset.filter("term", **{field: values})

        return queryset


class TermsFilter(QueryParamFilterBuilderMixin, FilterOptionsMixin, BaseFilterBackend):

    EQUAL = 'eq'
    NOT_EQUAL = 'neq'
    EXECUTION = 'op'

    valid_operators = [NOT_EQUAL, EXECUTION]
    real_operators = {EXECUTION: 'execution'}
    value_operators = [EQUAL, NOT_EQUAL]
    negation_operators = [NOT_EQUAL]

    real_operators_value_map = {
        'execution': {'any': 'plain', 'all': 'and'}
    }

    separator = getattr(api_settings, 'PARAM_VALUES_SEPARATOR', ',')

    def get_fields(self, view):
        return getattr(view, 'terms_filters', []) + self.get_option(view, 'fields', []) + list(self.get_option(view, 'field_map', {}).keys())

    def map_field(self, request, view, field):
        return self.get_option(view, 'field_map', {}).get(field, field)

    def map_value(self, request, view, field, operator, value):
        value_map = self.get_option(view, 'value_map', {})
        if field not in value_map:
            return value
        try:
            return value_map[field][value]
        except KeyError:
            raise ParseError("Invalid value for '{}'".format(field))

    def parse_value(self, request, view, field, operator, value):
        if operator in self.real_operators.values():
            try:
                return self.real_operators_value_map[operator][value]
            except KeyError:
                rev_op = dict((v, k) for k, v in self.real_operators.items())
                raise ParseError("Invalid value for modifier '{}'".format(rev_op.get(operator)))
        return [self.map_value(request, view, field, operator, i.strip()) for i in value.split(self.separator)]

    def build_filter(self, field, **params):
        return F("terms", **params)


class IndexedGeoShapeFilter(FilterOptionsMixin, BaseFilterBackend):

    def get_valid_params(self, view):
        return self.get_options(view).keys()

    def get_params(self, request, view):
        valid_params = self.get_valid_params(view)
        return {k: v for k, v in request.query_params.items() if k in valid_params}

    def get_filters(self, request, view):
        params = self.get_params(request, view)
        options = self.get_options(view)
        filters = []
        for param, value in params.iteritems():
            field_config = options.get(param)
            for id in value.split(','):
                filter_config = deepcopy(field_config['config'])
                field = field_config.get('field') if 'field' in field_config else param
                cache = field_config.get('cache') if 'cache' in field_config else False
                filter_config['indexed_shape']['id'] = id
                filters.append(F("geo_shape", **{field: filter_config, '_cache': cache}))
        return filters

    def filter_queryset(self, request, queryset, view):
        filters = self.get_filters(request, view)
        if filters:
            return queryset.filter("bool", must=[F("bool", should=filters)])
        return queryset


class MoreLikeThisFilter(FilterOptionsMixin, BaseFilterBackend):

    separator = getattr(api_settings, 'PARAM_VALUES_SEPARATOR', ',')
    LIKE_PARAM = getattr(api_settings, 'LIKE_PARAM', 'like')

    def get_like_ids(self, request):
        like_params = request.query_params.get(self.LIKE_PARAM)
        if like_params:
            return set([param.strip() for param in like_params.split(',')])
        return None

    def filter_queryset(self, request, queryset, view):

        like_params = self.get_like_ids(request)

        if like_params:

            options = self.get_options(view)

            docs = [
                {
                    "_type": options.get('type'),
                    "_id": id
                } for id in like_params
            ]

            queryset = queryset.query("more_like_this", fields=options.get('fields'), docs=docs,
                                      min_term_freq=options.get('min_term_freq'), max_query_terms=options.get('max_query_terms'))

        return queryset


class GeoDistanceFilter(FilterOptionsMixin, BaseFilterBackend):

    def get_valid_params(self, view):
        return self.get_option(view, 'fields', []) + list(self.get_option(view, 'field_map', {}).keys())

    def map_param_name(self, view, param_name):
        return self.get_option(view, 'field_map', {}).get(param_name, param_name)

    def parse_param_value(self, view, param_name, param_value):

        try:
            values = []
            for value in param_value.split(' '):
                matches = re.match(r'(\d+(mm|cm|m|km|in|yd|ft|mi|nmi))@([\-\+]?\d+(\.\d+)?),([\-\+]?\d+(\.\d+)?)', value)
                values.append((matches.group(1),) + (map(float, matches.group(5, 3)),))
            return values
        except:
            raise ParseError("Invalid distance for '{}'".format(param_name))

    def get_filters(self, request, view):
        valid_params = self.get_valid_params(view)
        params = {self.map_param_name(view, k): self.parse_param_value(view, k, v) for k, v in request.query_params.items() if k in valid_params}
        filters = []

        for field, values in params.items():
            for config in values:
                filters.append(F("geo_distance", **{field: config[1], 'distance': config[0]}))
        return filters

    def filter_queryset(self, request, queryset, view):
        filters = self.get_filters(request, view)
        if filters:
            return queryset.filter("bool", must=[F("bool", should=filters)])
        return queryset


class CannedTermsFilter(BaseFilterBackend):

    def get_valid_terms_filters(self, view):
        return getattr(view, 'canned_terms_filters', {})

    def get_terms_filters(self, request, valid_filters):
        return {valid_filters[k]['field']: valid_filters[k]['values'] for k, v in request.query_params.items() if k in valid_filters.keys() and bool(v)}

    def filter_queryset(self, request, queryset, view):
        valid_filters = self.get_valid_terms_filters(view)

        terms_filters = self.get_terms_filters(request, valid_filters)

        for field, values in terms_filters.items():
            queryset = queryset.filter("terms", **{field: values})

        return queryset


class IdsFilter(FilterOptionsMixin, BaseFilterBackend):

    separator = getattr(api_settings, 'PARAM_VALUES_SEPARATOR', ',')
    IDS_PARAM = getattr(api_settings, 'IDS_PARAM', 'id')

    def get_ids(self, view, request):
        type = self.get_option(view, 'type', None)
        id_params = request.query_params.get(self.IDS_PARAM)
        if id_params:
            return type, set([param.strip() for param in id_params.split(',')])
        return None, None

    def filter_queryset(self, request, queryset, view):
        type, ids = self.get_ids(view, request)
        if ids:
            return queryset.filter("ids", type=type, values=list(ids))
        return queryset


class MatchQueryFilter(BaseFilterBackend):

    def get_params(self, view):
        filters = getattr(view, 'match_query_filters', [])
        if isinstance(filters, six.string_types):
            return (filters,)
        return filters

    def get_match_query_filters(self, request, valid_filters):
        return {k: v.strip() for k, v in request.query_params.items() if k in valid_filters}

    def filter_queryset(self, request, queryset, view):
        valid_filters = self.get_params(view)

        query_filters = self.get_match_query_filters(request, valid_filters)

        for field, value in query_filters.items():
            queryset = queryset.filter("query", match={field: value})

        return queryset


class MatchQueriesFilter(BaseFilterBackend):

    separator = getattr(api_settings, 'PARAM_VALUES_SEPARATOR', ',')

    def get_params(self, view):
        filters = getattr(view, 'match_queries_filters', [])
        if isinstance(filters, six.string_types):
            return (filters,)
        return filters

    def get_match_queries_filters(self, request, valid_filters):
        return {k: [i.strip() for i in v.split(self.separator)] for k, v in request.query_params.items() if k in valid_filters}

    def filter_queryset(self, request, queryset, view):
        valid_filters = self.get_params(view)

        query_filters = self.get_match_queries_filters(request, valid_filters)

        should = None
        for field, values in query_filters.items():
            for value in values:
                if should is None:
                    should = F("query", match={field: value})
                else:
                    should |= F("query", match={field: value})
        if should:
            return queryset.filter("bool", must=F("bool", should))

        return queryset


class NumericRangeFilter(FilterOptionsMixin, BaseFilterBackend):

    GREATER_THAN = 'gt'
    GREATER_THAN_OR_EQUAL = 'gte'
    LESS_THAN = 'lt'
    LESS_THAN_OR_EQUAL = 'lte'
    TIME_ZONE = 'time_zone'

    OPERATOR_SYNONYMS = {
        "tz": TIME_ZONE,
    }

    VALID_SUFFIXES = [GREATER_THAN, GREATER_THAN_OR_EQUAL, LESS_THAN, LESS_THAN_OR_EQUAL, TIME_ZONE] + list(OPERATOR_SYNONYMS.keys())

    def get_valid_filters(self, view):
        filters = self.get_option(view, 'fields', []) + list(self.field_map.keys())
        if isinstance(filters, six.string_types):
            filters = (filters,)
        return ['{}.{}'.format(v, r) for v in filters for r in self.VALID_SUFFIXES]

    def get_range_query_filters(self, request, valid_filters):
        queries = defaultdict(dict)
        for k, v in ((k, v) for k, v in request.query_params.items() if k in valid_filters):
            field, operator = k.split('.')
            if operator not in [self.TIME_ZONE]:
                self.validate_field_value(field, v)
            if field in self.field_map:
                field = self.field_map[field]
            if operator in self.OPERATOR_SYNONYMS:
                operator = self.OPERATOR_SYNONYMS[operator]
            queries[field].update({operator: v})
        return queries

    def validate_field_value(self, field, value):
        validators = self.field_validators.get(field, [])
        for validator in validators:
            if validator == 'numeric':
                try:
                    float(value)
                except (ValueError, TypeError):
                    raise ParseError("Invalid numeric value for '{}'".format(field))
            elif validator == 'date':
                try:
                    parse(value)
                except (ValueError, TypeError):
                    raise ParseError("Invalid date value for '{}'".format(field))
            elif callable(validator):
                validator(value)

    def filter_queryset(self, request, queryset, view):
        self.field_map = self.get_option(view, 'field_map', {})
        self.field_validators = self.get_option(view, 'field_validators', {})
        valid_filters = self.get_valid_filters(view)
        query_filters = self.get_range_query_filters(request, valid_filters)
        for field, params in query_filters.items():
            queryset = queryset.filter("range", **{field: params})
        return queryset


class NumericBinFilter(FilterOptionsMixin, BaseFilterBackend):
    """
    Allow to filter by numeric range using a named bin

    Example::

        numeric_bin_filter_options = {
            "rank_level": {
                "1": {"rank": {"gte": 0, "lte": 20}},
                "2": {"rank": {"gte": 21, "lte": 40}},
                "3": {"rank": {"gte": 41, "lte": 60}},
                "4": {"rank": {"gte": 61, "lte": 80}},
                "5": {"rank": {"gte": 81, "lte": 100}},
            }
        }
    """

    def filter_queryset(self, request, queryset, view):
        bins = self.get_options(view)
        filters = []
        for bin_name, bin_options in bins.items():
            if bin_name in request.query_params:
                options = [v.strip() for v in request.query_params.get(bin_name).split(',')]
                for o in options:
                    try:
                        filters.append(F("range", **bin_options[o]))
                    except KeyError:
                        raise ParseError("Invalid value '{}' for '{}'".format(o, bin_name))
        if filters:
            return queryset.filter("bool", must=[F("bool", should=filters)])
        return queryset


class ExistsFilter(BaseFilterBackend):

    fields_param = getattr(api_settings, 'EXISTS_PARAM', 'exists')

    def get_fields(self, request):
        params = request.query_params.get(self.fields_param)
        if params:
            return set(param.strip() for param in params.split(','))

    def filter_queryset(self, request, queryset, view):
        if self.fields_param in request.query_params:
            for field in self.get_fields(request):
                queryset = queryset.filter("exists", field=field)
        return queryset


class MissingFilter(BaseFilterBackend):

    fields_param = getattr(api_settings, 'MISSING_PARAM', 'missing')

    def get_fields(self, request):
        params = request.query_params.get(self.fields_param)
        if params:
            return set(param.strip() for param in params.split(','))

    def filter_queryset(self, request, queryset, view):
        if self.fields_param in request.query_params:
            for field in self.get_fields(request):
                queryset = queryset.filter("missing", field=field)
        return queryset


class FieldsFilter(FilterOptionsMixin, BaseFilterBackend):

    fields_param = getattr(api_settings, 'FIELDS_PARAM', 'fields')

    def get_fields(self, view, request):
        valid = set(self.get_option(view, 'valid', []))
        default = set(self.get_option(view, 'default'))
        include = set(self.get_option(view, 'include', []))
        exclude = set(self.get_option(view, 'exclude', []))
        requested = set()
        params = request.query_params.get(self.fields_param)
        if params:
            requested = set([param.strip() for param in params.split(',')])
        fields = (include | (requested or default) - exclude)
        if valid:
            return fields & valid
        return fields

    def filter_queryset(self, request, queryset, view):
        fields = self.get_fields(view, request)

        if fields:
            return queryset.extra(_source=list(fields))
        return queryset


class TypesFilter(FilterOptionsMixin, BaseFilterBackend):

    fields_param = getattr(api_settings, 'FIELDS_PARAM', 'types')

    def get_types(self, view, request):
        valid = set(self.get_option(view, 'valid', []))
        default = set(self.get_option(view, 'default'))
        include = set(self.get_option(view, 'include', []))
        exclude = set(self.get_option(view, 'exclude', []))
        requested = set()
        params = request.query_params.get(self.fields_param)
        if params:
            requested = set([param.strip() for param in params.split(',')])
        types = (include | (requested or default) - exclude)
        if valid:
            return types & valid
        return types

    def filter_queryset(self, request, queryset, view):
        types = self.get_types(view, request)
        for type in types:
            queryset = queryset.filter("type", value=type)
        return queryset


class MinScoreFilter(BaseFilterBackend):

    min_score_param = getattr(api_settings, 'MIN_SCORE_PARAM', 'min_score')

    def get_min_score(self, request):
        min_score = request.query_params.get(self.min_score_param, None)
        if min_score:
            try:
                return float(min_score)
            except ValueError:
                raise ParseError("Invalid value for '{}'".format(self.min_score_param))

    def filter_queryset(self, request, queryset, view):
        min_score = self.get_min_score(request)
        if min_score:
            queryset = queryset.extra(min_score=min_score)
        return queryset


class OrderingFilter(FilterOptionsMixin, filters.OrderingFilter):
    """
    Same as default OrderingFilter, with identical view properties for ordering param and default ordering.

    Example::

        class YourView(APIView):

            ordering_filter_options = {
                "fields": ["id", "created", "updated", "start", "end", "rank", "duration", "country", "labels"],
                "field_map": {
                    "title": "title.raw"
                },
                "reversed": ["rank"],
                "default": ["-start"]
            }

    """
    special_reversed_fields = ["_score"]
    special_field_map = {"relevance": "_score"}

    def get_valid_fields(self, view):
        """
        Get fields that can be used to sort via the url

        """
        return self.get_option(view, 'fields', []) + self.get_option(view, 'field_map', {}).keys() + self.special_field_map.keys()

    def get_reversed_fields(self, view):
        """
        Get list of fields for which ordering should be reversed (- is ascending instead of descending, e.g. _score)
        """
        return self.get_option(view, 'reversed', []) + self.special_reversed_fields

    def get_field_map(self, view):
        """
        Get url name to storage name mapping

        """
        field_map = self.get_option(view, 'field_map', {})
        field_map.update(self.special_field_map)
        return field_map

    def get_ordering(self, request, queryset, view):
        """
        Prepare ordering directives based on valid fields, field map and reversed fields

        """
        params = request.query_params.get(self.ordering_param)

        if params:
            valid_fields = self.get_valid_fields(view)
            directives = (param.strip() for param in params.split(',') if param.strip().lstrip('-') in valid_fields)
        else:
            directives = self.get_default_ordering(view)

        if not directives:
            return

        field_map = self.get_field_map(view)
        reversed_fields = self.get_reversed_fields(view)
        ordering = []
        for directive in directives:
            name = directive.lstrip('-')
            field = field_map.get(name, name)
            if name in reversed_fields or field in reversed_fields:
                direction = 'asc' if directive.startswith('-') else 'desc'
            else:
                direction = 'desc' if directive.startswith('-') else 'asc'
            ordering.append({six.text_type(field): {six.text_type("order"): six.text_type(direction)}})
        return ordering


    def get_default_ordering(self, view):
        """
        Extract default ordering from ordering_filter_options.default, or use the view's ordering property

        """
        return self.get_option(view, 'default', None) or super(OrderingFilter, self).get_default_ordering(view)

    def filter_queryset(self, request, queryset, view):
        ordering = self.get_ordering(request, queryset, view)
        if ordering:
            queryset = queryset.sort(*ordering)
        return queryset



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from copy import copy
from rest_framework.renderers import JSONRenderer


class JSONPRenderer(JSONRenderer):
    """
    Renderer which serializes to json,
    wrapping the json output in a callback function.
    """

    media_type = 'application/javascript'
    format = 'jsonp'
    callback_parameter = 'callback'
    default_callback = 'callback'
    charset = 'utf-8'

    def get_callback(self, renderer_context):
        """
        Determine the name of the callback to wrap around the json output.
        """
        request = renderer_context.get('request', None)
        params = request and request.query_params or {}
        return params.get(self.callback_parameter, self.default_callback)

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Renders into jsonp, wrapping the json output in a callback function.

        Clients may set the callback function name using a query parameter
        on the URL, for example: ?callback=exampleCallbackName
        """
        renderer_context = renderer_context or {}
        callback = self.get_callback(renderer_context)
        json = super(JSONPRenderer, self).render(data, accepted_media_type,
                                                 renderer_context)
        return callback.encode(self.charset) + b'(' + json + b');'


class JSONContextDEBUGRenderer(JSONRenderer):
    """
    JSON Context Debug Render
    Includes Request Context in response for use in unit test assertions
    """

    def render(self, data, accepted_media_type=None, renderer_context=None):

        renderer_context = renderer_context or {}

        if isinstance(data, dict) and renderer_context:
            data['_context'] = {k: v for k, v in renderer_context.items() if k.startswith('_')}

        json = super(JSONContextDEBUGRenderer, self).render(data, accepted_media_type,
                                                 renderer_context)

        return json



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function

import inflection
from rest_framework import filters


class FilterOptionsMixin(object):

    view_options_defaults = {}

    @property
    def view_options_property(self):
        return "{}_options".format(inflection.underscore(self.__class__.__name__))

    def get_options(self, view):
        if not hasattr(self, '_options'):
            self._options = {}
            self._options.update(self.view_options_defaults)
            self._options.update(getattr(view, self.view_options_property, {}))
        return self._options

    def get_option(self, view, option, default=None):
        return self.get_options(view).get(option, default)


class MixedFilterBackend(filters.BaseFilterBackend):
    """
    Composed filters that add criteria based on whether a request is internal or external, from a user or a client

    Every access to a resource goes through this process:

            - if internal client: filter_internal_queryset
            - if external client: filter_external_queryset
            - if acting as user (App access): check user_permissions_set
            - if acting as client (API access): filter_client_queryset
                - if internal client: filter_internal_client_queryset
                - if external client: filter_external_client_queryset

    It's important to note that:

        * No filtering is done by default
        * Any of the methods can be overridden to provide specific behaviour

    """
    def filter_queryset(self, request, queryset, view):
        if hasattr(request, 'client') and request.client.is_internal:
            queryset = self.filter_internal_queryset(request, queryset, view)
        else:
            queryset = self.filter_external_queryset(request, queryset, view)

        if request.user.is_authenticated():
            return self.filter_user_queryset(request, queryset, view)
        else:
            return self.filter_client_queryset(request, queryset, view)

    def filter_internal_queryset(self, request, queryset, view):
        return queryset

    def filter_external_queryset(self, request, queryset, view):
        return queryset

    def filter_user_queryset(self, request, queryset, view):
        return queryset

    def filter_client_queryset(self, request, queryset, view):
        if request.client.is_internal:
            return self.filter_internal_client_queryset(request, queryset, view)
        else:
            return self.filter_external_client_queryset(request, queryset, view)

    def filter_internal_client_queryset(self, request, queryset, view):
        return queryset

    def filter_external_client_queryset(self, request, queryset, view):
        return queryset



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rfhq.mixins import NestedViewMixin


class RoutingAPIMixin(object):

    def get_routing_key(self):
        raise NotImplementedError()

    def get_serializer_context(self):
        ctx = super(RoutingAPIMixin, self).get_serializer_context()
        ctx.update({"routing_key": self.get_routing_key()})
        return ctx


class TermNestedViewMixin(NestedViewMixin):

    def filter_nested_queryset(self, queryset):
        parents_query_dict = self.get_parents_kwargs()
        if parents_query_dict:
            return queryset.filter("term", **parents_query_dict)
        else:
            return queryset


class TermNestedViewSetMixin(TermNestedViewMixin):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.conf.urls import url, include

from rest_framework.routers import DefaultRouter, SimpleRouter, Route, DynamicListRoute, DynamicDetailRoute


class EndpointRouter(SimpleRouter):
    routes = [
        # List route.
        Route(
            url=r'^{prefix}$',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            name='{basename}-list',
            initkwargs={'suffix': 'List'}
        ),
        # Dynamically generated list routes.
        # Generated using @list_route decorator
        # on methods of the viewset.
        DynamicListRoute(
            url=r'^{prefix}{methodname}{trailing_slash}$',
            name='{basename}-{methodnamehyphen}',
            initkwargs={}
        ),
        # Detail route.
        Route(
            url=r'^{prefix}{lookup}{trailing_slash}$',
            mapping={
                'get': 'retrieve',
                'put': 'update',
                'patch': 'partial_update',
                'delete': 'destroy'
            },
            name='{basename}-detail',
            initkwargs={'suffix': 'Instance'}
        ),
        # Dynamically generated detail routes.
        # Generated using @detail_route decorator on methods of the viewset.
        DynamicDetailRoute(
            url=r'^{prefix}{lookup}/{methodname}{trailing_slash}$',
            name='{basename}-{methodnamehyphen}',
            initkwargs={}
        ),
    ]

class NestedRegistryItem(object):
    """
    From djangorestframework-extensions

    """
    def __init__(self, router, parent_prefix, parent_item=None):
        self.router = router
        self.parent_prefix = parent_prefix
        self.parent_item = parent_item

    def register(self, prefix, viewset, base_name, parents_query_lookups):
        self.router._register(
            prefix=self.get_prefix(current_prefix=prefix, parents_query_lookups=parents_query_lookups),
            viewset=viewset,
            base_name=base_name,
        )
        return NestedRegistryItem(
            router=self.router,
            parent_prefix=prefix,
            parent_item=self
        )

    def register_urls(self, prefix, urlpatterns, base_name, parents_query_lookups):
        self.router.register_urls(self.get_prefix(current_prefix=prefix, parents_query_lookups=parents_query_lookups), urlpatterns, base_name)

    def get_prefix(self, current_prefix, parents_query_lookups):
        return '{0}/{1}'.format(
            self.get_parent_prefix(parents_query_lookups),
            current_prefix
        )

    def get_parent_prefix(self, parents_query_lookups):
        prefix = '/'
        current_item = self
        i = len(parents_query_lookups) - 1
        while current_item:
            prefix = '{parent_prefix}/(?P<{parent_pk_kwarg_name}>[^/.]+)/{prefix}'.format(
                parent_prefix=current_item.parent_prefix,
                parent_pk_kwarg_name='parent_lookup_{}'.format(parents_query_lookups[i]),
                prefix=prefix
            )
            i -= 1
            current_item = current_item.parent_item
        return prefix.strip('/')


class NestableRouterMixin(object):
    """
    From django-rest-framework-extensions (but removed obsolete code that broke with newer versions of django-rest-framework)

    """
    def __init__(self, *args, **kwargs):
        super(NestableRouterMixin, self).__init__()
        self._includes = []

    def _register(self, *args, **kwargs):
        return super(NestableRouterMixin, self).register(*args, **kwargs)

    def register(self, *args, **kwargs):
        self._register(*args, **kwargs)
        return NestedRegistryItem(router=self, parent_prefix=self.registry[-1][0])

    def register_urls(self, prefix, urlpatterns, base_name):
        self._includes.append((prefix, urlpatterns, base_name))

    def get_urls(self):
        urls = super(NestableRouterMixin, self).get_urls()
        for prefix, urlpatterns, base_name in self._includes:
            urls.append(url(''.join([prefix, self.trailing_slash]), include(urlpatterns), name=base_name))
        return urls

class SimpleNestedRouter(NestableRouterMixin, SimpleRouter):

    pass


class DefaultNestedRouter(NestableRouterMixin, DefaultRouter):

    pass



from django.core.exceptions import ImproperlyConfigured
from django.http.request import QueryDict
from django.utils.decorators import classonlymethod
from django.utils.functional import cached_property
import re
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework import status

from .serializers import LenientBulkCreateSerializerMixin


class LenientBulkCreateViewMixin(object):

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if not isinstance(serializer, LenientBulkCreateSerializerMixin):
            raise ImproperlyConfigured("serializer_class must inherit from 'LenientBulkCreateSerializer'")
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        if serializer.bulk_validation_errors:
            # Return description of item and errors but accept the rest
            return Response(serializer.bulk_validation_errors, status=status.HTTP_202_ACCEPTED, headers=headers)
        # Return empty list as it can be massively long
        return Response([], status=status.HTTP_201_CREATED, headers=headers)


class BlendedRequest(Request):

    def __init__(self, *args, **kwargs):
        params_prefix = kwargs.pop('params_prefix')
        self._params_prefix = '{}.'.format(params_prefix) if params_prefix and not params_prefix.endswith('.') else ''
        self._mapped_params = kwargs.pop('mapped_params', {}) or {}
        super(BlendedRequest, self).__init__(*args, **kwargs)

    @cached_property
    def query_params(self):
        orignal_params = super(BlendedRequest, self).query_params
        if self._params_prefix:
            params = QueryDict(mutable=True)
            params.update({k[len(self._params_prefix):]: v for k, v in orignal_params.items() if k.startswith(self._params_prefix)})
        else:
            params = super(BlendedRequest, self).query_params
        for m_k, m_v in  self._mapped_params.items():
            for o_k, o_v in orignal_params.items():
                matches = re.match(m_v, o_k)
                if matches:
                    params[''.join([m_k, (matches.group(1) if matches.lastindex >= 1 else '')])] = o_v
        return params


class BlendableViewMixin(object):

    @classonlymethod
    def blend(cls, request, params_prefix=None, mapped_params=None, *args, **kwargs):
        self = cls()
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.request = self.initialize_blended_request(request, params_prefix, mapped_params, *args, **kwargs)
        return self

    def initialize_blended_request(self, request, params_prefix, mapped_params, *args, **kwargs):
        return BlendedRequest(
            request,
            parsers=self.get_parsers(),
            authenticators=self.get_authenticators(),
            negotiator=self.get_content_negotiator(),
            parser_context=self.get_parser_context(request),
            params_prefix=params_prefix,
            mapped_params=mapped_params
        )

    def get_filtered_queryset(self):
        return self.filter_queryset(self.get_queryset())


class SelfLookupMixin(object):
    """
    Allows looking up objects using the 'self' keyword in urls in place of the primary key for the currently logged in user/account.

    When 'self' is found as the lookup value, it is substituted by the result of the call to get_self_lookup_value.

    When using nested routers, each key containing self is substituted by the result of the call to get_{key}_self_lookup_value.

    """
    self_pk = 'self'

    def get_self_lookup_value(self):
        raise NotImplementedError("View must implement get_self_lookup_value()")

    def initial(self, request, *args, **kwargs):
        super(SelfLookupMixin, self).initial(request, *args, **kwargs)
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg in kwargs and kwargs[lookup_url_kwarg] == self.self_pk:
            self.kwargs[lookup_url_kwarg] = kwargs[lookup_url_kwarg] = self.get_self_lookup_value()

    def get_parents_kwargs(self):
        result = super(SelfLookupMixin, self).get_parents_kwargs()
        for query_lookup, query_value in result.items():
            if query_value == self.self_pk:
                if not hasattr(self, 'get_{}_self_lookup_value'.format(query_lookup)):
                        raise NotImplementedError("View must implement get_{}_self_lookup_value()".format(query_lookup))
                query_value = getattr(self, 'get_{}_self_lookup_value'.format(query_lookup))()
                result[query_lookup] = query_value
        return result


class MixedSerializerMixin(object):
    """
    Allows defining a different serializer on the view for internal and external clients

    """
    internal_serializer_class = None
    external_serializer_class = None

    def get_serializer_class(self):
        if hasattr(self.request, 'client') and self.request.client.is_internal:
            return self.get_internal_serializer()
        else:
            return self.get_external_serializer()

    def get_internal_serializer(self):
        return self.internal_serializer_class

    def get_external_serializer(self):
        return self.external_serializer_class


class NestedViewMixin(object):
    """
    Allows nesting a viewset under another viewset for filtered access

    From djangorestframework-extensions

    """
    def get_queryset(self):
        return self.filter_nested_queryset(super(NestedViewMixin, self).get_queryset())

    def filter_nested_queryset(self, queryset):
        parents_query_dict = self.get_parents_kwargs()
        if parents_query_dict:
            return queryset.filter(**parents_query_dict)
        else:
            return queryset

    def get_parents_kwarg(self, key, default=None):
        return self.get_parents_kwargs().get(key, default)

    def get_parents_kwargs(self):
        if hasattr(self, '_parent_kwargs'):
            return self._parent_kwargs
        result = {}
        for kwarg_name in self.kwargs:
            if kwarg_name.startswith('parent_lookup_'):
                query_lookup = kwarg_name.replace('parent_lookup_', '', 1)
                query_value = self.kwargs.get(kwarg_name)
                result[query_lookup] = query_value
        self._parent_kwargs = result
        return result

    def is_nested(self, parent_lookup_key=None):
        if parent_lookup_key:
            return "parent_lookup_{}".format(parent_lookup_key) in self.kwargs
        return any(kwarg_name.startswith('parent_lookup_') for kwarg_name in self.kwargs)


class NestedViewSetMixin(NestedViewMixin):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rest_framework.parsers import BaseParser


class SkipParser(BaseParser):

    media_type = '*/*'

    def parse(self, stream, media_type=None, parser_context=None):
        return


def detail_route_view(methods=None, **kwargs):
    """
    Used to mark a method on a ViewSet that should be routed to a different view for detail requests.

    The decorator must ensure that the ViewSet's permissions, parsers, method restrictions, etc. are ignored
    """
    view_kwargs = kwargs
    fake_kwargs = {'permission_classes': (), 'parser_classes': (SkipParser,), 'url_path': view_kwargs.pop('url_path', None), 'throttle_classes': []}
    methods = ['get', 'post', 'put', 'patch', 'options', 'head', 'delete'] if (methods is None) else methods

    def decorator(func=None):
        def deferred_view(self, request, *args, **kwargs):
            if not hasattr(func, 'view'):
                # instantiate view once only
                func.view = func(self).as_view(**view_kwargs)
            return func.view(request._request, *args, **kwargs)
        deferred_view.bind_to_methods = methods
        deferred_view.detail = True
        deferred_view.kwargs = fake_kwargs
        return deferred_view
    return decorator



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.apps import apps

from django.utils import six
from django.utils.translation import ugettext_lazy as _

from rest_framework.serializers import Serializer, ListSerializer, ModelField


class SourceSerializer(Serializer):

    def to_representation(self, instance):
        return instance['_source']


class RoutedDocTypeMixin(object):

    def get_routing_key(self):
        return self.context.get("routing_key", None)


class DocTypeSerializer(RoutedDocTypeMixin, Serializer):

    def to_representation(self, instance):
        data = super(DocTypeSerializer, self).to_representation(instance)
        data.update(instance.to_dict())
        return data


class DocTypeListSerializer(ListSerializer):

    def to_representation(self, instance):
        return instance.to_dict()


class AggregationsSerializer(Serializer):

    pass


class TermsAggregationField(Serializer):

    def to_representation(self, instance):
        return [{"name": d['key'], "count": d['doc_count']} for d in instance['buckets']]


class DateHistogramAggregationField(Serializer):

    def to_bucket_representation(self, bucket):
        return {"date": bucket['key_as_string'], "count": bucket['doc_count']}

    def to_representation(self, instance):
        return (self.to_bucket_representation(bucket) for bucket in instance['buckets'])


class GeoHashAggregationField(Serializer):

    def to_representation(self, instance):
        return (self.to_bucket_representation(bucket) for bucket in instance['buckets'])


class StatsAggregationField(Serializer):

    def to_representation(self, instance):
        return instance.to_dict()


class GeoBoundsAggregationField(Serializer):

    def to_representation(self, instance):
        bounds = getattr(instance, 'bounds', None)
        if bounds:
            return [bounds['top_left']['lon'], bounds['top_left']['lat']] + [bounds['bottom_right']['lon'], bounds['bottom_right']['lat']]
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function

from rest_framework.views import get_view_name as get_default_view_name


def get_view_name(view_cls, suffix=None):
    if hasattr(view_cls, 'view_name'):
        return view_cls.view_name
    return get_default_view_name(view_cls, suffix)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from collections import OrderedDict
from django.core.exceptions import ImproperlyConfigured
from elasticsearch.exceptions import TransportError
from rest_framework.exceptions import ParseError

from rest_framework.pagination import _get_count, LimitOffsetPagination, BasePagination, CursorPagination
from rest_framework.response import Response


class ElasticsearchPagination(LimitOffsetPagination):

    max_limit = 200

    def paginate_queryset(self, queryset, request, view=None):
        self.limit = self.get_limit(request)
        if self.limit is None:
            return None
        self.offset = self.get_offset(request)
        # try:
        results = queryset[self.offset:self.offset + self.limit].execute()
        # except TransportError:
            # raise ParseError("Invalid query or parameter value")
        self.count = results.hits.total
        if 'suggest' in results:
            self.suggestions = results.suggest.to_dict()
        else:
            self.suggestions = []
        if 'aggregations' in results:
            self.aggregations = results.aggregations.to_dict()
        else:
            self.aggregations = {}
        self.request = request
        if self.count > self.limit and self.template is not None:
            self.display_page_controls = True
        return results

    def get_paginated_response(self, data):
        items = [
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data),
        ]
        if self.suggestions:
            items.append(('suggestions', self.suggestions))
        if self.aggregations:
            items.append(('aggregations', self.aggregations))
        return Response(OrderedDict(items))

# Not used for now (only returns aggregations as the results)
# class ElasticsearchAggregationsPagination(LimitOffsetPagination):
#
#     display_page_controls = False
#
#     def paginate_queryset(self, queryset, request, view=None):
#         results = queryset.params(search_type="count").execute()
#         data = results.aggregations
#         self.count = None
#         self.request = request
#         return data
#
#     def get_paginated_response(self, data):
#         items = [
#             ('count', self.count),
#             ('next', None),
#             ('previous', None),
#             ('results', data),
#         ]
#         return Response(OrderedDict(items))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import os
import codecs
import re
import uuid
from setuptools import setup, find_packages
from pip.req import parse_requirements

def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='rfhq',
    version=find_version('rfhq', '__init__.py'),
    author='PredictHQ',
    long_description=read('README.md'),
    install_requires=[str(ir.req)
                      for ir in parse_requirements('requirements.txt', session=uuid.uuid4())],
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Framework :: Django",
    ]
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function

import logging
from django.utils.translation import ugettext_lazy as _
from rest_framework.exceptions import APIException
from rest_framework import status
from rest_framework.views import exception_handler as default_exception_handler
from rest_framework.response import Response

log = logging.getLogger(__name__)

def exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = default_exception_handler(exc, context)

    if response is None:
        # Capture Oauthlib exceptions
        try:
            # noinspection PyUnresolvedReferences
            from oauthlib.oauth2.rfc6749.errors import OAuth2Error
            if isinstance(exc, OAuth2Error):
                response = Response({'detail': exc.error}, status=exc.status_code)
        except:
            pass

    # Now add the HTTP status code to the response.
    if response is not None:
        if isinstance(response.data, dict):
            response.data['code'] = response.status_code
            response.data['error'] = response.data.pop('detail', None)
        elif isinstance(response.data, list):
            response.data = {'error': response.data.pop(), 'code': response.status_code}

    if response:
        log.error("Framework exception ({}) occurred: {}".format(response.status_code, exc))

    return response


class APIUnavailable(APIException):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE
    default_detail = _('Service temporarily unavailable')


class ResourceGone(APIException):
    status_code = status.HTTP_410_GONE
    default_detail = _('No longer available')

class InvalidSession(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _("Your Session has expired, you will need to login again")